// Copyright (c) 2016, Tobias Mueller tm(at)tm3d.de
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the
//    distribution.
//  * All advertising materials mentioning features or use of this
//    software must display the following acknowledgement: This product
//    includes software developed by tm3d.de and its contributors.
//  * Neither the name of tm3d.de nor the names of its contributors may
//    be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <unistd.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <time.h>
#include <termios.h>
#include <errno.h>
#include <sys/time.h>
#include "owCOMInterface.h"

// Mode Commands
#define MODE_DATA                      0xE1
#define MODE_COMMAND                   0xE3
#define MODE_STOP_PULSE                0xF1

// Return byte value
#define RB_CHIPID_MASK                 0x1C
#define RB_RESET_MASK                  0x03
#define RB_1WIRESHORT                  0x00
#define RB_PRESENCE                    0x01
#define RB_ALARMPRESENCE               0x02
#define RB_NOPRESENCE                  0x03

#define RB_BIT_MASK                    0x03
#define RB_BIT_ONE                     0x03
#define RB_BIT_ZERO                    0x00

// Masks for all bit ranges
#define CMD_MASK                       0x80
#define FUNCTSEL_MASK                  0x60
#define BITPOL_MASK                    0x10
#define SPEEDSEL_MASK                  0x0C
#define MODSEL_MASK                    0x02
#define PARMSEL_MASK                   0x70
#define PARMSET_MASK                   0x0E
#define VERSION_MASK                   0x1C

// Command or config bit
#define CMD_COMM                       0x81
#define CMD_CONFIG                     0x01

// Function select bits
#define FUNCTSEL_BIT                   0x00
#define FUNCTSEL_SEARCHON              0x30
#define FUNCTSEL_SEARCHOFF             0x20
#define FUNCTSEL_RESET                 0x40
#define FUNCTSEL_CHMOD                 0x60

// Bit polarity/Pulse voltage bits
#define BITPOL_ONE                     0x10
#define BITPOL_ZERO                    0x00
#define BITPOL_5V                      0x00
#define BITPOL_12V                     0x10

// One Wire speed bits
#define SPEEDSEL_STD                   0x00
#define SPEEDSEL_FLEX                  0x04
#define SPEEDSEL_OD                    0x08
#define SPEEDSEL_PULSE                 0x0C

// Data/Command mode select bits
#define MODSEL_DATA                    0x00
#define MODSEL_COMMAND                 0x02

// 5V Follow Pulse select bits (If 5V pulse
// will be following the next byte or bit.)
#define PRIME5V_TRUE                   0x02
#define PRIME5V_FALSE                  0x00

// Parameter select bits
#define PARMSEL_PARMREAD               0x00
#define PARMSEL_SLEW                   0x10
#define PARMSEL_12VPULSE               0x20
#define PARMSEL_5VPULSE                0x30
#define PARMSEL_WRITE1LOW              0x40
#define PARMSEL_SAMPLEOFFSET           0x50
#define PARMSEL_ACTIVEPULLUPTIME       0x60
#define PARMSEL_BAUDRATE               0x70

// Pull down slew rate.
#define PARMSET_Slew15Vus              0x00
#define PARMSET_Slew2p2Vus             0x02
#define PARMSET_Slew1p65Vus            0x04
#define PARMSET_Slew1p37Vus            0x06
#define PARMSET_Slew1p1Vus             0x08
#define PARMSET_Slew0p83Vus            0x0A
#define PARMSET_Slew0p7Vus             0x0C
#define PARMSET_Slew0p55Vus            0x0E

// 12V programming pulse time table
#define PARMSET_32us                   0x00
#define PARMSET_64us                   0x02
#define PARMSET_128us                  0x04
#define PARMSET_256us                  0x06
#define PARMSET_512us                  0x08
#define PARMSET_1024us                 0x0A
#define PARMSET_2048us                 0x0C
#define PARMSET_infinite               0x0E

// 5V strong pull up pulse time table
#define PARMSET_16p4ms                 0x00
#define PARMSET_65p5ms                 0x02
#define PARMSET_131ms                  0x04
#define PARMSET_262ms                  0x06
#define PARMSET_524ms                  0x08
#define PARMSET_1p05s                  0x0A
#define PARMSET_2p10s                  0x0C
#define PARMSET_infinite               0x0E

// Write 1 low time
#define PARMSET_Write8us               0x00
#define PARMSET_Write9us               0x02
#define PARMSET_Write10us              0x04
#define PARMSET_Write11us              0x06
#define PARMSET_Write12us              0x08
#define PARMSET_Write13us              0x0A
#define PARMSET_Write14us              0x0C
#define PARMSET_Write15us              0x0E

// Data sample offset and Write 0 recovery time
#define PARMSET_SampOff3us             0x00
#define PARMSET_SampOff4us             0x02
#define PARMSET_SampOff5us             0x04
#define PARMSET_SampOff6us             0x06
#define PARMSET_SampOff7us             0x08
#define PARMSET_SampOff8us             0x0A
#define PARMSET_SampOff9us             0x0C
#define PARMSET_SampOff10us            0x0E

// Active pull up on time
#define PARMSET_PullUp0p0us            0x00
#define PARMSET_PullUp0p5us            0x02
#define PARMSET_PullUp1p0us            0x04
#define PARMSET_PullUp1p5us            0x06
#define PARMSET_PullUp2p0us            0x08
#define PARMSET_PullUp2p5us            0x0A
#define PARMSET_PullUp3p0us            0x0C
#define PARMSET_PullUp3p5us            0x0E

// Baud rate bits
#define PARMSET_9600                   0x00
#define PARMSET_19200                  0x02
#define PARMSET_57600                  0x04
#define PARMSET_115200                 0x06

// DS2480B program voltage available
#define DS2480PROG_MASK                0x20

// mode bit flags
#define MODE_NORMAL                    0x00
#define MODE_OVERDRIVE                 0x01
#define MODE_STRONG5                   0x02
#define MODE_PROGRAM                   0x04
#define MODE_BREAK                     0x08

// Versions of DS2480
#define VER_LINK                       0x1C
#define VER_DS2480                     0x08
#define VER_DS2480B                    0x0C




//---------------------------------------------------------------------------
//  Description:
//     flush the rx and tx buffers
//
// 'portnum'  - number 0 to MAX_PORTNUM-1.  This number was provided to
//              OpenCOM to indicate the port number.
//
void owCOMInterface::FlushCOM() {
   tcflush(fd, TCIOFLUSH);
}


int owCOMInterface::OpenCOM(uint8_t comnr) {
	struct termios t;               // see man termios - declared as above
	int rc;
	//int fd;
	char port_zstr[100];
	if (com_init) return fd;

	sprintf(port_zstr,"/dev/ttyUSB%i",comnr);

   fd = open(port_zstr, O_RDWR|O_NONBLOCK);
   if (fd<0)
   {
	   log->set(OWLOG_ERROR,"ERROR open Com %s return %i",port_zstr,fd);
      return fd;
   }
   rc = tcgetattr (fd, &t);
   if (rc < 0)
   {
      int tmp;
      tmp = errno;
      close(fd);
      errno = tmp;
      log->set(OWLOG_ERROR,"OWERROR_SYSTEM_RESOURCE_INIT_FAILED %s",port_zstr);
      return rc; // changed (2.00), used to return rc;
   }

   cfsetospeed(&t, B9600);
   cfsetispeed (&t, B9600);

   // Get terminal parameters. (2.00) removed raw
   tcgetattr(fd,&t);
   // Save original settings.
   origterm = t;

   // Set to non-canonical mode, and no RTS/CTS handshaking
   t.c_iflag &= ~(BRKINT|ICRNL|IGNCR|INLCR|INPCK|ISTRIP|IXON|IXOFF|PARMRK);
   t.c_iflag |= IGNBRK|IGNPAR;
   t.c_oflag &= ~(OPOST);
   t.c_cflag &= ~(CRTSCTS|CSIZE|HUPCL|PARENB);
   t.c_cflag |= (CLOCAL|CS8|CREAD);
   t.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ICANON|IEXTEN|ISIG);
   t.c_cc[VMIN] = 0;
   t.c_cc[VTIME] = 3;

   rc = tcsetattr(fd, TCSAFLUSH, &t);
   tcflush(fd,TCIOFLUSH);

   if (rc < 0)
   {
      int tmp;
      tmp = errno;
      close(fd);
      errno = tmp;
      log->set(OWLOG_ERROR,"OWERROR_SYSTEM_RESOURCE_INIT_FAILED %s",port_zstr);
      return rc; // changed (2.00), used to return rc;
   }
	com_init=1;
   return fd; // changed (2.00), used to return fd;
}


//---------------------------------------------------------------------------
// Closes the connection to the port.
//
// 'portnum'  - number 0 to MAX_PORTNUM-1.  This number was provided to
//              OpenCOM to indicate the port number.
//
void owCOMInterface::CloseCOM()
{
   // restore tty settings
   tcsetattr(fd, TCSAFLUSH, &origterm);
   FlushCOM();
   close(fd);
   com_init=0;
   
}


//--------------------------------------------------------------------------
// Write an array of bytes to the COM port, verify that it was
// sent out.  Assume that baud rate has been set.
//
// 'portnum'   - number 0 to MAX_PORTNUM-1.  This number provided will
//               be used to indicate the port number desired when calling
//               all other functions in this library.
// Returns 1 for success and 0 for failure
//
int owCOMInterface::WriteCOM( int outlen, uint8_t *outbuf)
{
   long count = outlen;
   int i = write(fd, outbuf, outlen);

   tcdrain(fd);
   return (i == count);
}


//--------------------------------------------------------------------------
// Read an array of bytes to the COM port, verify that it was
// sent out.  Assume that baud rate has been set.
//
// 'portnum'  - number 0 to MAX_PORTNUM-1.  This number was provided to
//              OpenCOM to indicate the port number.
// 'outlen'   - number of bytes to write to COM port
// 'outbuf'   - pointer ot an array of bytes to write
//
// Returns:  TRUE(1)  - success
//           FALSE(0) - failure
//
int owCOMInterface::ReadCOM( int inlen, uint8_t *inbuf)
{
   fd_set         filedescr;
   struct timeval tval;
   int            cnt;

   // loop to wait until each byte is available and read it
   for (cnt = 0; cnt < inlen; cnt++)
   {
      // set a descriptor to wait for a character available
      FD_ZERO(&filedescr);
      FD_SET(fd,&filedescr);
      // set timeout to 10ms
      tval.tv_sec = 0;
      tval.tv_usec = 10000;

      // if byte available read or return bytes read
      if (select(fd+1,&filedescr,NULL,NULL,&tval) != 0)
      {
         if (read(fd,&inbuf[cnt],1) != 1) {
			log->set(OWLOG_ERROR,"Read Error on Serial");

            return cnt;
         }
      }
      else {
		log->set(OWLOG_ERROR,"Read Error on Serial (select)");
        return cnt;
	}
   }

   // success, so return desired length
   return inlen;
}




//--------------------------------------------------------------------------
//  Description:
//     Send a break on the com port for at least 2 ms
//
// 'portnum'  - number 0 to MAX_PORTNUM-1.  This number was provided to
//              OpenCOM to indicate the port number.
//
void owCOMInterface::BreakCOM()
{
   int duration = 0;              // see man termios break may be
   tcsendbreak(fd, duration);     // too long
}


//--------------------------------------------------------------------------
// Set the baud rate on the com port.
//
// 'portnum'   - number 0 to MAX_PORTNUM-1.  This number was provided to
//               OpenCOM to indicate the port number.
// 'new_baud'  - new baud rate defined as
// PARMSET_9600     0x00
// PARMSET_19200    0x02
// PARMSET_57600    0x04
// PARMSET_115200   0x06
//
void owCOMInterface::SetBaudCOM( uint8_t new_baud)
{
   struct termios t;
   int rc;
   speed_t baud;

   // read the attribute structure
   rc = tcgetattr(fd, &t);
   if (rc < 0)
   {
      close(fd); 
      log->set(OWLOG_ERROR,"Error on Serial (set Boudrate)");

      return;
   }

   // convert parameter to linux baud rate
   switch(new_baud)
   {
      case PARMSET_9600:
         baud = B9600;
         break;
      case PARMSET_19200:
         baud = B19200;
         break;
      case PARMSET_57600:
         baud = B57600;
         break;
      case PARMSET_115200:
         baud = B115200;
         break;
      default:
         baud = B9600;
         break;
   }

   // set baud in structure
   cfsetospeed(&t, baud);
   cfsetispeed(&t, baud);

   // change baud on port
   rc = tcsetattr(fd, TCSAFLUSH, &t);
   if (rc < 0) {
	  log->set(OWLOG_ERROR,"Error on Serial (set Boudrate)");

      close(fd);
  }
}


//---------------------------------------------------------------------------
// Attempt to resyc and detect a DS2480B
//
// 'portnum'    - number 0 to MAX_PORTNUM-1.  This number was provided to
//                OpenCOM to indicate the port number.
//
// Returns:  TRUE  - DS2480B detected successfully
//           FALSE - Could not detect DS2480B
//

int8_t owCOMInterface::DS2480Detect(){
   uint8_t sendpacket[10],readbuffer[10];
   uint8_t sendlen=0;

   // reset modes
   UMode = MODSEL_COMMAND;
   UBaud = PARMSET_9600;
   USpeed = SPEEDSEL_FLEX;

   // set the baud rate to 9600
   SetBaudCOM((uint8_t)UBaud);

   // send a break to reset the DS2480
   BreakCOM();

   // delay to let line settle
	usleep(2000);

   // flush the buffers
   FlushCOM();

   // send the timing byte
   sendpacket[0] = 0xC1;
   if (WriteCOM(1,sendpacket) != 1)
   {
		log->set(OWLOG_ERROR,"Error on Serial (Write in DS2480Detect)");
      return FALSE;
   }

   // delay to let line settle
   usleep(4000);

   // set the FLEX configuration parameters
   // default PDSRC = 1.37Vus
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_SLEW | PARMSET_Slew1p37Vus;
   // default W1LT = 10us
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_WRITE1LOW | PARMSET_Write10us;
   // default DSO/WORT = 8us
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_SAMPLEOFFSET | PARMSET_SampOff8us;

   // construct the command to read the baud rate (to test command block)
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_PARMREAD | (PARMSEL_BAUDRATE >> 3);

   // also do 1 bit operation (to test 1-Wire block)
   sendpacket[sendlen++] = CMD_COMM | FUNCTSEL_BIT | UBaud | BITPOL_ONE;

   // flush the buffers
   FlushCOM();

   // send the packet
   if (WriteCOM(sendlen,sendpacket))
   {
      // read back the response
      if (ReadCOM(5,readbuffer) == 5)
      {
         // look at the baud rate and bit operation
         // to see if the response makes sense
         if (((readbuffer[3] & 0xF1) == 0x00) &&
             ((readbuffer[3] & 0x0E) == UBaud) &&
             ((readbuffer[4] & 0xF0) == 0x90) &&
             ((readbuffer[4] & 0x0C) == UBaud))
            return TRUE;
         else
			log->set(OWLOG_ERROR,"OWERROR_DS2480_BAD_RESPONSE\n");
      }
      else
		//	printf("OWERROR_READCOM_FAILED\n");
         log->set(OWLOG_ERROR,"OWERROR_READCOM_FAILED");
   }
   else
	//printf("OWERROR_READCOM_FAILED\n");
      log->set(OWLOG_ERROR,"OWERROR_WRITECOM_FAILED");

   return FALSE;
}


















int owCOMInterface::InitAdapter(uint8_t nr) {
	// attempt to open the communications port
	if ((fd = OpenCOM(nr)) < 0)
	{
		log->set(OWLOG_ERROR,"OWERROR_OPENCOM_FAILED");
		return -1;
	}

	// detect DS2480
	if (!DS2480Detect())
	{
		CloseCOM();
		log->set(OWLOG_ERROR,"OWERROR_DS2480_NOT_DETECTED");
		return -1;
	}
	com_init=1;
	return 1;	
}

void owCOMInterface::ReleaseAdapter() {
	CloseCOM();
}


int owCOMInterface::Reset() {
	uint8_t readbuffer[10],sendpacket[10];
	uint8_t sendlen=0;

	// check if correct mode
	if (UMode != MODSEL_COMMAND) {
		UMode = MODSEL_COMMAND;
		sendpacket[sendlen++] = MODE_COMMAND;
	}

	// construct the command
	sendpacket[sendlen++] = (uint8_t)(CMD_COMM | FUNCTSEL_RESET | USpeed);
	// flush the buffers
	FlushCOM();
	// send the packet
	if (WriteCOM(sendlen,sendpacket))
	{
		// read back the 1 byte response
		if (ReadCOM(1,readbuffer) == 1)
		{
			// make sure this byte looks like a reset byte
			if (((readbuffer[0] & RB_RESET_MASK) == RB_PRESENCE) ||
				((readbuffer[0] & RB_RESET_MASK) == RB_ALARMPRESENCE))
			{
            // check if programming voltage available
            UVersion = (readbuffer[0] & VERSION_MASK);
            return 1;
         }
         else
           log->set(OWLOG_ERROR,"OWERROR_RESET_FAILED");

      }
     else
       log->set(OWLOG_ERROR,"OWERROR_READCOM_FAILED");
   }
   else
     log->set(OWLOG_ERROR,"OWERROR_WRITECOM_FAILED");

   // an error occured so re-sync with DS2480
   DS2480Detect();

   return FALSE;	
	
}

uint8_t owCOMInterface::sendrecivBit(uint8_t bit) { 
	uint8_t readbuffer[10],sendpacket[10];
	uint8_t sendlen=0;

	// check if correct mode
	if (UMode != MODSEL_COMMAND) {
		UMode = MODSEL_COMMAND;
		sendpacket[sendlen++] = MODE_COMMAND;
	}

	// construct the command
	sendpacket[sendlen] = (bit != 0) ? BITPOL_ONE : BITPOL_ZERO;
	sendpacket[sendlen++] |= CMD_COMM | FUNCTSEL_BIT | USpeed;
	// flush the buffers
	FlushCOM();

	// send the packet
	if (WriteCOM(sendlen,sendpacket)) {
    // read back the response
		if (ReadCOM(1,readbuffer) == 1) {
			// interpret the response
			if (((readbuffer[0] & 0xE0) == 0x80) &&
				((readbuffer[0] & RB_BIT_MASK) == RB_BIT_ONE))
				return 1;
			else
				return 0;
		}
      else
         log->set(OWLOG_ERROR,"OWERROR_READCOM_FAILED");
	} else
      log->set(OWLOG_ERROR,"OWERROR_WRITECOM_FAILED");

   // an error occured so re-sync with DS2480
	DS2480Detect();
	return 0;
}


uint8_t owCOMInterface::sendrecivByte(uint8_t byte) {
   uint8_t readbuffer[10],sendpacket[10];
   uint8_t sendlen=0;
	
	
   // check if correct mode
   if (UMode != MODSEL_DATA)
   {
      UMode = MODSEL_DATA;
      sendpacket[sendlen++] = MODE_DATA;
   }

   // add the byte to send
   sendpacket[sendlen++] = byte;

   // check for duplication of data that looks like COMMAND mode
   if (byte ==(uint8_t)MODE_COMMAND)
      sendpacket[sendlen++] = (uint8_t)byte;

   // flush the buffers
   FlushCOM();

   // send the packet
   if (WriteCOM(sendlen,sendpacket))
   {
      // read back the 1 byte response
      if (ReadCOM(1,readbuffer) == 1)
      {

          // return the response
          return readbuffer[0];
      }
      else
         log->set(OWLOG_ERROR,"OWERROR_READCOM_FAILED");
   }
   else
      log->set(OWLOG_ERROR,"OWERROR_WRITECOM_FAILED");

   // an error occured so re-sync with DS2480
   DS2480Detect();
   return 0;
}

	
